/*
 * Copyright (c) 1998-2001 groomed <groomed@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package pharerouge.hotline;

import java.io.DataInputStream;
import java.io.IOException;

import pharerouge.utils.DebuggerOutput;

/**
 *  Reads events from the server and dispatches new threads to handle them.
 */
class ReaderThread extends Thread {

    HLClient hlc;
    DataInputStream input;
    boolean blocked = false;
    HLProtocol.Packet packet;

    /**
     *  Constructor for the ReaderThread object
     *
     * @param  h   Description of the Parameter
     * @param  in  Description of the Parameter
     */
    ReaderThread(HLClient h, DataInputStream in) {
        super();
        hlc = h;
        input = in;
        DebuggerOutput.debug("ReaderThread[" + this + "] created");
    }

    /**
     *  This is the main input loop where all incoming data is read and
     *  dispatched.
     */
    public void run() {
        DebuggerOutput.debug("ReaderThread[" + this + "]: running");

        /* 2002.04.21 MackyBee added experimentally */
        String reason = null;
        /* 2002.04.25 MackyBee added */
        boolean terminate = false;

        try {
            while (!this.isInterrupted()) {

                honourBlock();
                packet = hlc.hlp.new Packet(input);

                DebuggerOutput.debug("ReaderThread: got packet " +
                    packet.toString());

                if (packet.header.id == HLProtocol.HTLS_HDR_TASK) {
                    if (hlc.lockTask(packet.header.trans)) {
                        HLTask t = hlc.getTask(packet.header.trans);
                        if (t.type == HLProtocol.HTLC_HDR_FILE_GET ||
                            t.type == HLProtocol.HTLC_HDR_FILE_PUT) {
                            DebuggerOutput.debug("ReaderThread[" + this 
                                + "]: task[" + packet.header.trans 
                                + "]: " 
                                + "dispatching file transfer in separate "
                                + "thread.");
                            FileTransferThread ft = new FileTransferThread(
                                hlc, packet);
                            ft.start();
                        } else {
                            DebuggerOutput.debug("ReaderThread[" + this 
                                + "]: task[" + packet.header.trans 
                                + "]: dispatching...");
                            hlc.hlr.dispatch(packet);
                        }
                    } else {
                        DebuggerOutput.debug("ReaderThread[" + this + 
                            "]: task[" + packet.header.trans 
                            + "]: cannot lock, will not dispatch.");
                    }
                } else {
                    DebuggerOutput.debug("ReaderThread[" + this 
                        + "]: dispatching...");
                    hlc.hlr.dispatch(packet);
                }
            }

        // 2003.01.30 MackyBee modified. 
        // Like "disconnected" dialog message should be displayed. So added. 
        // 2002.04.21 modified, 2002.04.18 MackyBee modified.
        // Almost SocketException are usual. On now program structure,
        // when connection was closed, this exception is always occurred. */
        } catch (IOException ioe) {
            terminate = true;
            // 2003.01.31 MackyBee added this if-statement. 
            if (!(this.isInterrupted())) {
                reason = ioe.getMessage();
                new Error("The connection with "
                    + hlc.getServerName() 
                    + " has been deactivated.");
            }
        /*} catch (SocketException se) {
            terminate = true;
            reason = se.getMessage();
            // No problem, maybe
            DebuggerOutput.debug("ReaderThread.run: " + reason);
            // 2002.04.18 MackyBee added.
        } catch (EOFException ee) {
            terminate = true;
            reason = ee.getMessage();
            DebuggerOutput.debug("ReaderThread.run: " + reason);*/
        } catch (Exception e) {
            terminate = true;
            reason = e.getMessage();

            // 2002.09.29 MackyBee modified.
            e.printStackTrace();
            //DebuggerOutput.stackTrace(e);

            //{{{ deactivated old lines
            /* 2002.04.21 MackyBee disabled experimentally */
            //String reason = e.toString();
            //if (e.getMessage() != null) {
            //    reason = e.getMessage();
            //}
            //hlc.terminateAllTasks(reason);

            /* 2002.04.21 MackyBee moved this to under */
            /* If we were explicitly interrupted, then we do not want
               to fire an unexpected disconnect event. */
            //if (!isInterrupted()) {
            //    DebuggerOutput.debug("ReaderThread[" + this +
            //        "]: calling fireDisconnect ...");
            //    hlc.fireDisconnect(reason);
            //}}}

        /* 2002.04.25 modified.
           There is in case "reason == null" if it should be disconnected.
           2002.04.21 MackyBee moved this from above. And modified. */
        } finally {
            if (terminate) {
                //if (reason != null) {
                    hlc.terminateAllTasks(reason);
                    if (!isInterrupted()) {
                        DebuggerOutput.debug("ReaderThread[" + this +
                            "]: calling fireDisconnect ...");
                        hlc.fireDisconnect(reason);
                    }
                //}
            }
        }

        DebuggerOutput.debug("ReaderThread[" + this + "] exiting");
    }

    /**
     *  This method controls whether event listeners are called or not. If the
     *  argument is true, then event listeners will not be called until
     *  setBlocked is invoked again with a false argument.
     *
     * @param  b  true to block, false to unblock
     */
    synchronized void setBlocked(boolean b) {
        blocked = b;

        if (blocked == false) {
            notify();
        }
    }

    /**
     *  Blocks until {@link #setBlocked} is called with a false parameter.
     *
     * @exception  InterruptedException  Description of the Exception
     */
    // 2002.10.21 MackyBee modified synchronized treat.
    void honourBlock() throws InterruptedException {
        if (blocked) {
            DebuggerOutput.debug("ReaderThread[" + this 
                + "]: waiting for unblock");
            synchronized (this) {
                while (blocked) {
                    // 2002.10.21 MackyBee modified experimentally.
                    wait();
                    //wait(5000);
                }
            }
        }
    }
}
